@@ -132,6 +132,8 @@ static NS_DEFINE_CID(kXULControllersCID,  NS_XULCONTROLLERS_CID);
                                         ? ((bitfield) |=  (0x01 << (field))) \
                                         : ((bitfield) &= ~(0x01 << (field))))
 
+static const char* kWhitespace = "\n\r\t\b";
+
 class nsHTMLInputElement : public nsGenericHTMLFormElement,
                            public nsImageLoadingContent,
                            public nsIDOMHTMLInputElement,
@@ -257,11 +259,6 @@ protected:
                             nsITextControlFrame* aFrame);
 
   nsresult GetSelectionRange(PRInt32* aSelectionStart, PRInt32* aSelectionEnd);
-  //Helper method
-#ifdef ACCESSIBILITY
-  nsresult FireEventForAccessibility(nsPresContext* aPresContext,
-                                     const nsAString& aEventType);
-#endif
 
   /**
    * Get the name if it exists and return whether it did exist
@@ -352,6 +349,13 @@ protected:
   nsAutoPtr<nsString>      mFileName;
 };
 
+#ifdef ACCESSIBILITY
+//Helper method
+static nsresult FireEventForAccessibility(nsIDOMHTMLInputElement* aTarget,
+                                          nsPresContext* aPresContext,
+                                          const nsAString& aEventType);
+#endif
+
 //
 // construction, destruction
 //
@@ -370,6 +374,7 @@ nsHTMLInputElement::nsHTMLInputElement(nsINodeInfo *aNodeInfo,
 
 nsHTMLInputElement::~nsHTMLInputElement()
 {
+  DestroyImageLoadingContent();
   if (mValue) {
     nsMemory::Free(mValue);
   }
@@ -392,6 +397,7 @@ NS_HTML_CONTENT_INTERFACE_MAP_BEGIN(nsHTMLInputElement,
   NS_INTERFACE_MAP_ENTRY(nsIRadioControlElement)
   NS_INTERFACE_MAP_ENTRY(nsIPhonetic)
   NS_INTERFACE_MAP_ENTRY(imgIDecoderObserver)
+  NS_INTERFACE_MAP_ENTRY(imgIDecoderObserver_MOZILLA_1_8_BRANCH)
   NS_INTERFACE_MAP_ENTRY(nsIImageLoadingContent)
   NS_INTERFACE_MAP_ENTRY(nsIDOMNSEditableElement)
   NS_INTERFACE_MAP_ENTRY_CONTENT_CLASSINFO(HTMLInputElement)
@@ -583,7 +589,7 @@ nsHTMLInputElement::GetForm(nsIDOMHTMLFormElement** aForm)
   return nsGenericHTMLFormElement::GetForm(aForm);
 }
 
-NS_IMPL_STRING_ATTR(nsHTMLInputElement, DefaultValue, value)
+//NS_IMPL_STRING_ATTR(nsHTMLInputElement, DefaultValue, value)
 NS_IMPL_BOOL_ATTR(nsHTMLInputElement, DefaultChecked, checked)
 NS_IMPL_STRING_ATTR(nsHTMLInputElement, Accept, accept)
 NS_IMPL_STRING_ATTR(nsHTMLInputElement, AccessKey, accesskey)
@@ -601,6 +607,25 @@ NS_IMPL_STRING_ATTR(nsHTMLInputElement, UseMap, usemap)
 //NS_IMPL_INT_ATTR_DEFAULT_VALUE(nsHTMLInputElement, Size, size, 0)
 //NS_IMPL_STRING_ATTR_DEFAULT_VALUE(nsHTMLInputElement, Type, type, "text")
 
+NS_IMETHODIMP
+nsHTMLInputElement::GetDefaultValue(nsAString& aValue)
+{
+  GetAttrHelper(nsHTMLAtoms::value, aValue);
+
+  if (mType != NS_FORM_INPUT_HIDDEN) {
+    // Bug 114997: trim \n, etc. for non-hidden inputs
+    aValue = nsContentUtils::TrimCharsInSet(kWhitespace, aValue);
+  }
+
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsHTMLInputElement::SetDefaultValue(const nsAString& aValue)
+{
+  return SetAttrHelper(nsHTMLAtoms::value, aValue);
+}
+
 NS_IMETHODIMP
 nsHTMLInputElement::GetSize(PRUint32* aValue)
 {
@@ -673,6 +698,10 @@ nsHTMLInputElement::GetValue(nsAString& aValue)
   // Treat value == defaultValue for other input elements
   nsresult rv = GetAttr(kNameSpaceID_None, nsHTMLAtoms::value, aValue);
 
+  if (mType != NS_FORM_INPUT_HIDDEN) {
+    aValue = nsContentUtils::TrimCharsInSet(kWhitespace, aValue);
+  }
+
   if (rv == NS_CONTENT_ATTR_NOT_THERE &&
       (mType == NS_FORM_INPUT_RADIO || mType == NS_FORM_INPUT_CHECKBOX)) {
     // The default value of a radio or checkbox input is "on".
@@ -1155,13 +1184,15 @@ nsHTMLInputElement::SetFocus(nsPresContext* aPresContext)
     do_QueryInterface(doc->GetScriptGlobalObject());
   if (win) {
     nsIFocusController *focusController = win->GetRootFocusController();
-    PRBool isActive = PR_FALSE;
-    focusController->GetActive(&isActive);
-    if (!isActive) {
-      focusController->SetFocusedWindow(win);
-      focusController->SetFocusedElement(this);
+    if (focusController) {
+      PRBool isActive = PR_FALSE;
+      focusController->GetActive(&isActive);
+      if (!isActive) {
+        focusController->SetFocusedWindow(win);
+        focusController->SetFocusedElement(this);
 
-      return;
+        return;
+      }
     }
   }
 
@@ -1207,13 +1238,15 @@ nsHTMLInputElement::Select()
       do_QueryInterface(doc->GetScriptGlobalObject());
     if (win) {
       nsIFocusController *focusController = win->GetRootFocusController();
-      PRBool isActive = PR_FALSE;
-      focusController->GetActive(&isActive);
-      if (!isActive) {
-        focusController->SetFocusedWindow(win);
-        focusController->SetFocusedElement(this);
-        SelectAll(presContext);
-        return NS_OK;
+      if (focusController) {
+        PRBool isActive = PR_FALSE;
+        focusController->GetActive(&isActive);
+        if (!isActive) {
+          focusController->SetFocusedWindow(win);
+          focusController->SetFocusedElement(this);
+          SelectAll(presContext);
+          return NS_OK;
+        }
       }
     }
 
@@ -1471,6 +1504,18 @@ nsHTMLInputElement::HandleDOMEvent(nsPresContext* aPresContext,
   //
   PRInt32 oldType = mType;
 
+  // Fire onchange (if necessary), before we do the blur, bug 357684.
+  if (aEvent->message == NS_BLUR_CONTENT) {
+    nsIFrame* primaryFrame = GetPrimaryFrame(PR_FALSE);
+    if (primaryFrame) {
+      nsITextControlFrame* textFrame = nsnull;
+      CallQueryInterface(primaryFrame, &textFrame);
+      if (textFrame) {
+        textFrame->CheckFireOnChange();
+      }
+    }
+  }
+
   // Try script event handlers first if its not a focus/blur event
   //we dont want the doc to get these
   rv = nsGenericHTMLFormElement::HandleDOMEvent(aPresContext, aEvent,
@@ -1542,10 +1587,13 @@ nsHTMLInputElement::HandleDOMEvent(nsPresContext* aPresContext,
 #ifdef ACCESSIBILITY
       // Fire an event to notify accessibility
       if (mType == NS_FORM_INPUT_CHECKBOX) {
-        FireEventForAccessibility(aPresContext,
+        FireEventForAccessibility(this, aPresContext,
                                   NS_LITERAL_STRING("CheckboxStateChange"));
       } else {
-        FireEventForAccessibility(aPresContext,
+        FireEventForAccessibility(this, aPresContext,
+                                  NS_LITERAL_STRING("RadioStateChange"));
+        // Fire event for the previous selected radio.
+        FireEventForAccessibility(selectedRadioButton, aPresContext,
                                   NS_LITERAL_STRING("RadioStateChange"));
       }
 #endif
@@ -1714,14 +1762,8 @@ nsHTMLInputElement::HandleDOMEvent(nsPresContext* aPresContext,
           if (mType == NS_FORM_INPUT_BUTTON || 
               mType == NS_FORM_INPUT_RESET || 
               mType == NS_FORM_INPUT_SUBMIT ) {
-            nsCOMPtr<nsIDOMNSEvent> nsevent;
-
             if (aDOMEvent) {
-              nsevent = do_QueryInterface(*aDOMEvent);
-            }
-
-            if (nsevent) {
-              nsevent->PreventBubble();
+              (*aDOMEvent)->StopPropagation();
             } else {
               rv = NS_ERROR_FAILURE;
             }
@@ -2145,13 +2187,17 @@ nsHTMLInputElement::GetPhonetic(nsAString& aPhonetic)
 }
 
 #ifdef ACCESSIBILITY
-nsresult
-nsHTMLInputElement::FireEventForAccessibility(nsPresContext* aPresContext,
-                                              const nsAString& aEventType)
+/*static*/ nsresult
+FireEventForAccessibility(nsIDOMHTMLInputElement* aTarget,
+                          nsPresContext* aPresContext,
+                          const nsAString& aEventType)
 {
+  nsCOMPtr<nsIContent> content(do_QueryInterface(aTarget));
+  NS_ENSURE_TRUE(content, NS_ERROR_FAILURE);
+
   nsCOMPtr<nsIDOMEvent> event;
   nsCOMPtr<nsIEventListenerManager> manager;
-  GetListenerManager(getter_AddRefs(manager));
+  content->GetListenerManager(getter_AddRefs(manager));
   if (manager &&
       NS_SUCCEEDED(manager->CreateEvent(aPresContext, nsnull,
                                         NS_LITERAL_STRING("Events"),
@@ -2163,9 +2209,8 @@ nsHTMLInputElement::FireEventForAccessibility(nsPresContext* aPresContext,
       privateEvent->SetTrusted(PR_TRUE);
     }
 
-    nsISupports *target = NS_STATIC_CAST(nsIDOMHTMLInputElement*, this);
     PRBool defaultActionEnabled;
-    aPresContext->EventStateManager()->DispatchNewEvent(target, event,
+    aPresContext->EventStateManager()->DispatchNewEvent(aTarget, event,
                                                         &defaultActionEnabled);
   }
 
@@ -2432,9 +2477,7 @@ nsHTMLInputElement::SubmitNamesValues(nsIFormSubmission* aFormSubmission,
   // Get the value
   nsAutoString value;
   rv = GetValue(value);
-  if (NS_FAILED(rv)) {
-    return rv;
-  }
+  NS_ENSURE_SUCCESS(rv, rv);
 
   if (mType == NS_FORM_INPUT_SUBMIT && value.IsEmpty() &&
       !HasAttr(kNameSpaceID_None, nsHTMLAtoms::value)) {
