@@ -277,17 +277,17 @@ XULPopupListenerImpl::PreLaunchPopup(nsIDOMEvent* aMouseEvent)
   // Get the document with the popup.
   nsCOMPtr<nsIContent> content = do_QueryInterface(mElement);
 
-  // Turn the document into a XUL document so we can use SetPopupNode.
-  nsCOMPtr<nsIDOMXULDocument> xulDocument = do_QueryInterface(content->GetDocument());
+  // Turn the document into a XUL document so we can use SetPopupNode
+  nsCOMPtr<nsIDOMXULDocument2> xulDocument = do_QueryInterface(content->GetDocument());
   if (!xulDocument) {
     NS_ERROR("Popup attached to an element that isn't in XUL!");
     return NS_ERROR_FAILURE;
   }
 
   // Store clicked-on node in xul document for context menus and menu popups.
+  // CLEAR THE POPUP EVENT BEFORE THIS FUNCTION EXITS
   xulDocument->SetPopupNode( targetNode );
-
-  nsCOMPtr<nsIDOMNSEvent> nsevent(do_QueryInterface(aMouseEvent));
+  xulDocument->SetTrustedPopupEvent( aMouseEvent );
 
   switch (popupType) {
     case eXULPopupType_popup:
@@ -296,11 +296,7 @@ XULPopupListenerImpl::PreLaunchPopup(nsIDOMEvent* aMouseEvent)
       if (button == 0) {
         // Time to launch a popup menu.
         LaunchPopup(aMouseEvent);
-
-        if (nsevent) {
-            nsevent->PreventBubble();
-        }
-
+        aMouseEvent->StopPropagation();
         aMouseEvent->PreventDefault();
       }
       break;
@@ -313,14 +309,11 @@ XULPopupListenerImpl::PreLaunchPopup(nsIDOMEvent* aMouseEvent)
     FireFocusOnTargetContent(targetNode);
 #endif
     LaunchPopup(aMouseEvent);
-
-    if (nsevent) {
-        nsevent->PreventBubble();
-    }
-
+    aMouseEvent->StopPropagation();
     aMouseEvent->PreventDefault();
     break;
   }
+  xulDocument->SetTrustedPopupEvent(nsnull);
   return NS_OK;
 }
 
