@@ -1,4 +1,5 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
+ * vim: set ts=8 sw=4 et tw=78:
  *
  * ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
@@ -278,14 +279,18 @@ DropWatchPoint(JSContext *cx, JSWatchPoint *wp)
 }
 
 void
-js_MarkWatchPoints(JSRuntime *rt)
+js_MarkWatchPoints(JSContext *cx)
 {
+    JSRuntime *rt;
     JSWatchPoint *wp;
 
+    rt = cx->runtime;
     for (wp = (JSWatchPoint *)rt->watchPointList.next;
          wp != (JSWatchPoint *)&rt->watchPointList;
          wp = (JSWatchPoint *)wp->links.next) {
-        MARK_SCOPE_PROPERTY(wp->sprop);
+        MARK_SCOPE_PROPERTY(cx, wp->sprop);
+        if (wp->sprop->attrs & JSPROP_SETTER)
+            JS_MarkGCThing(cx, wp->setter, "wp->setter", NULL);
     }
 }
 
@@ -360,18 +365,67 @@ js_watch_set(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
                              vp, wp->closure);
             if (ok) {
                 /*
-                 * Create pseudo-frame for call to setter so that any
-                 * stack-walking security code in the setter will correctly
-                 * identify the guilty party.
+                 * Create a pseudo-frame for the setter invocation so that any
+                 * stack-walking security code under the setter will correctly
+                 * identify the guilty party.  So that the watcher appears to
+                 * be active to obj_eval and other such code, point frame.pc
+                 * at the JSOP_STOP at the end of the script.
                  */
-                JSObject *funobj = (JSObject *) wp->closure;
-                JSFunction *fun = (JSFunction *) JS_GetPrivate(cx, funobj);
+                JSObject *closure;
+                JSClass *clasp;
+                JSFunction *fun;
+                JSScript *script;
+                uintN nslots;
+                jsval smallv[5];
+                jsval *argv;
                 JSStackFrame frame;
 
+                closure = (JSObject *) wp->closure;
+                clasp = OBJ_GET_CLASS(cx, closure);
+                if (clasp == &js_FunctionClass) {
+                    fun = (JSFunction *) JS_GetPrivate(cx, closure);
+                    script = FUN_SCRIPT(fun);
+                } else if (clasp == &js_ScriptClass) {
+                    fun = NULL;
+                    script = (JSScript *) JS_GetPrivate(cx, closure);
+                } else {
+                    fun = NULL;
+                    script = NULL;
+                }
+
+                nslots = 2;
+                if (fun) {
+                    nslots += fun->nargs;
+                    if (FUN_NATIVE(fun))
+                        nslots += fun->u.n.extra;
+                }
+
+                if (nslots <= JS_ARRAY_LENGTH(smallv)) {
+                    argv = smallv;
+                } else {
+                    argv = JS_malloc(cx, nslots * sizeof(jsval));
+                    if (!argv) {
+                        DropWatchPoint(cx, wp);
+                        return JS_FALSE;
+                    }
+                }
+
+                argv[0] = OBJECT_TO_JSVAL(closure);
+                argv[1] = JSVAL_NULL;
+                memset(argv + 2, 0, (nslots - 2) * sizeof(jsval));
+
                 memset(&frame, 0, sizeof(frame));
-                frame.script = FUN_SCRIPT(fun);
+                frame.script = script;
+                if (script) {
+                    JS_ASSERT(script->length >= JSOP_STOP_LENGTH);
+                    frame.pc = script->code + script->length
+                             - JSOP_STOP_LENGTH;
+                }
                 frame.fun = fun;
+                frame.argv = argv + 2;
                 frame.down = cx->fp;
+                frame.scopeChain = OBJ_GET_PARENT(cx, closure);
+
                 cx->fp = &frame;
                 ok = !wp->setter ||
                      ((sprop->attrs & JSPROP_SETTER)
@@ -379,8 +433,10 @@ js_watch_set(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
                                         1, vp, vp)
                       : wp->setter(cx, OBJ_THIS_OBJECT(cx, obj), userid, vp));
                 cx->fp = frame.down;
+                if (argv != smallv)
+                    JS_free(cx, argv);
             }
-            return DropWatchPoint(cx, wp);
+            return DropWatchPoint(cx, wp) && ok;
         }
     }
     JS_ASSERT(0);       /* XXX can't happen */
@@ -396,6 +452,7 @@ js_watch_set_wrapper(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
     jsval userid;
 
     funobj = JSVAL_TO_OBJECT(argv[-2]);
+    JS_ASSERT(OBJ_GET_CLASS(cx, funobj) == &js_FunctionClass);
     wrapper = (JSFunction *) JS_GetPrivate(cx, funobj);
     userid = ATOM_KEY(wrapper->atom);
     *rval = argv[0];
@@ -535,10 +592,8 @@ JS_SetWatchPoint(JSContext *cx, JSObject *obj, jsval id,
             JS_free(cx, wp);
             goto out;
         }
-        JS_APPEND_LINK(&wp->links, &rt->watchPointList);
         wp->object = obj;
-        wp->sprop = sprop;
-        JS_ASSERT(sprop->setter != js_watch_set);
+        JS_ASSERT(sprop->setter != js_watch_set || pobj != obj);
         wp->setter = sprop->setter;
         wp->nrefs = 1;
 
@@ -546,10 +601,16 @@ JS_SetWatchPoint(JSContext *cx, JSObject *obj, jsval id,
         sprop = js_ChangeNativePropertyAttrs(cx, obj, sprop, 0, sprop->attrs,
                                              sprop->getter, watcher);
         if (!sprop) {
+            /* Self-link wp->links so DropWatchPoint can JS_REMOVE_LINK it. */
+            JS_INIT_CLIST(&wp->links);
             DropWatchPoint(cx, wp);
             ok = JS_FALSE;
             goto out;
         }
+        wp->sprop = sprop;
+
+        /* Now that wp is fully initialized, append it to rt's wp list. */
+        JS_APPEND_LINK(&wp->links, &rt->watchPointList);
     }
     wp->handler = handler;
     wp->closure = closure;
@@ -782,7 +843,7 @@ JS_GetFrameScopeChain(JSContext *cx, JSStackFrame *fp)
 {
     /* Force creation of argument and call objects if not yet created */
     (void) JS_GetFrameCallObject(cx, fp);
-    return fp->scopeChain;
+    return js_GetScopeChain(cx, fp);
 }
 
 JS_PUBLIC_API(JSObject *)
@@ -790,19 +851,15 @@ JS_GetFrameCallObject(JSContext *cx, JSStackFrame *fp)
 {
     if (! fp->fun)
         return NULL;
-#if JS_HAS_ARGS_OBJECT
+
     /* Force creation of argument object if not yet created */
     (void) js_GetArgsObject(cx, fp);
-#endif
-#if JS_HAS_CALL_OBJECT
+
     /*
      * XXX ill-defined: null return here means error was reported, unlike a
      *     null returned above or in the #else
      */
     return js_GetCallObject(cx, fp, NULL);
-#else
-    return NULL;
-#endif /* JS_HAS_CALL_OBJECT */
 }
 
 
@@ -901,14 +958,19 @@ JS_SetDestroyScriptHook(JSRuntime *rt, JSDestroyScriptHook hook,
 
 JS_PUBLIC_API(JSBool)
 JS_EvaluateUCInStackFrame(JSContext *cx, JSStackFrame *fp,
-                          const jschar *bytes, uintN length,
+                          const jschar *chars, uintN length,
                           const char *filename, uintN lineno,
                           jsval *rval)
 {
+    JSObject *scobj;
     uint32 flags, options;
     JSScript *script;
     JSBool ok;
 
+    scobj = JS_GetFrameScopeChain(cx, fp);
+    if (!scobj)
+        return JS_FALSE;
+
     /*
      * XXX Hack around ancient compiler API to propagate the JSFRAME_SPECIAL
      * flags to the code generator (see js_EmitTree's TOK_SEMI case).
@@ -917,16 +979,16 @@ JS_EvaluateUCInStackFrame(JSContext *cx, JSStackFrame *fp,
     fp->flags |= JSFRAME_DEBUGGER | JSFRAME_EVAL;
     options = cx->options;
     cx->options = options | JSOPTION_COMPILE_N_GO;
-    script = JS_CompileUCScriptForPrincipals(cx, fp->scopeChain,
+    script = JS_CompileUCScriptForPrincipals(cx, scobj,
                                              JS_StackFramePrincipals(cx, fp),
-                                             bytes, length, filename, lineno);
+                                             chars, length, filename, lineno);
     fp->flags = flags;
     cx->options = options;
     if (!script)
         return JS_FALSE;
 
-    ok = js_Execute(cx, fp->scopeChain, script, fp,
-                    JSFRAME_DEBUGGER | JSFRAME_EVAL, rval);
+    ok = js_Execute(cx, scobj, script, fp, JSFRAME_DEBUGGER | JSFRAME_EVAL,
+                    rval);
     js_DestroyScript(cx, script);
     return ok;
 }
@@ -939,10 +1001,12 @@ JS_EvaluateInStackFrame(JSContext *cx, JSStackFrame *fp,
 {
     jschar *chars;
     JSBool ok;
+    size_t len = length;
 
-    chars = js_InflateString(cx, bytes, length);
+    chars = js_InflateString(cx, bytes, &len);
     if (!chars)
         return JS_FALSE;
+    length = (uintN) len;
     ok = JS_EvaluateUCInStackFrame(cx, fp, chars, length, filename, lineno,
                                    rval);
     JS_free(cx, chars);
@@ -1023,12 +1087,10 @@ JS_GetPropertyDesc(JSContext *cx, JSObject *obj, JSScopeProperty *sprop,
     pd->flags |= ((sprop->attrs & JSPROP_ENUMERATE) ? JSPD_ENUMERATE : 0)
               | ((sprop->attrs & JSPROP_READONLY)  ? JSPD_READONLY  : 0)
               | ((sprop->attrs & JSPROP_PERMANENT) ? JSPD_PERMANENT : 0)
-#if JS_HAS_CALL_OBJECT
               | ((getter == js_GetCallVariable)    ? JSPD_VARIABLE  : 0)
-#endif /* JS_HAS_CALL_OBJECT */
               | ((getter == js_GetArgument)        ? JSPD_ARGUMENT  : 0)
               | ((getter == js_GetLocalVariable)   ? JSPD_VARIABLE  : 0);
-#if JS_HAS_CALL_OBJECT
+
     /* for Call Object 'real' getter isn't passed in to us */
     if (OBJ_GET_CLASS(cx, obj) == &js_CallClass &&
         getter == js_CallClass.getProperty) {
@@ -1043,7 +1105,7 @@ JS_GetPropertyDesc(JSContext *cx, JSObject *obj, JSScopeProperty *sprop,
                      ? JSPD_ARGUMENT
                      : JSPD_VARIABLE;
     }
-#endif /* JS_HAS_CALL_OBJECT */
+
     pd->spare = 0;
     pd->slot = (pd->flags & (JSPD_ARGUMENT | JSPD_VARIABLE))
                ? sprop->shortid
@@ -1232,24 +1294,15 @@ GetAtomTotalSize(JSContext *cx, JSAtom *atom)
 JS_PUBLIC_API(size_t)
 JS_GetFunctionTotalSize(JSContext *cx, JSFunction *fun)
 {
-    size_t nbytes, obytes;
-    JSObject *obj;
-    JSAtom *atom;
+    size_t nbytes;
 
     nbytes = sizeof *fun;
-    JS_ASSERT(fun->nrefs);
-    obj = fun->object;
-    if (obj) {
-        obytes = JS_GetObjectTotalSize(cx, obj);
-        if (fun->nrefs > 1)
-            obytes = JS_HOWMANY(obytes, fun->nrefs);
-        nbytes += obytes;
-    }
-    if (fun->interpreted)
-        nbytes += JS_GetScriptTotalSize(cx, fun->u.script);
-    atom = fun->atom;
-    if (atom)
-        nbytes += GetAtomTotalSize(cx, atom);
+    if (fun->object)
+        nbytes += JS_GetObjectTotalSize(cx, fun->object);
+    if (FUN_INTERPRETED(fun))
+        nbytes += JS_GetScriptTotalSize(cx, fun->u.i.script);
+    if (fun->atom)
+        nbytes += GetAtomTotalSize(cx, fun->atom);
     return nbytes;
 }
 
@@ -1322,7 +1375,7 @@ JS_GetScriptFilenameFlags(JSScript *script)
     JS_ASSERT(script);
     if (!script->filename)
         return JSFILENAME_NULL;
-    return js_GetScriptFilenameFlags(script->filename);    
+    return js_GetScriptFilenameFlags(script->filename);
 }
 
 JS_PUBLIC_API(JSBool)
@@ -1343,7 +1396,7 @@ JS_PUBLIC_API(void)
 JS_FlagSystemObject(JSContext *cx, JSObject *obj)
 {
     uint8 *flagp;
-    
+
     flagp = js_GetGCThingFlags(obj);
     *flagp |= GCF_SYSTEM;
 }
