@@ -23,11 +23,12 @@
  * Contributor(s):
  *   timeless
  *   slucy@objectivesw.co.uk
- *   Hkan Waara <hwaara@chello.se>
+ *   Håkan Waara <hwaara@chello.se>
  *   Jan Varga <varga@ku.sk>
  *   Seth Spitzer <sspitzer@netscape.com>
  *   David Bienvenu <bienvenu@netscape.com>
  *   Ian Neal <bugzilla@arlen.demon.co.uk>
+ *   Karsten Düsterloh <mnyromyr@tprac.de>
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either of the GNU General Public License Version 2 or later (the "GPL"),
@@ -65,8 +66,8 @@ const kAllowRemoteContent = 2;
 const kIsAPhishMessage = 0;
 const kNotAPhishMessage = 1;
 
-const kMsgNotificationJunkBar = 1;
-const kMsgNotificationPhishingBar = 2;
+const kMsgNotificationPhishingBar = 1;
+const kMsgNotificationJunkBar = 2;
 const kMsgNotificationRemoteImages = 3;
 
 var gMessengerBundle;
@@ -228,7 +229,7 @@ function InitViewSortByMenu()
     setSortByMenuItemCheckState("sortByStatusMenuitem", (sortType == nsMsgViewSortType.byStatus));
     setSortByMenuItemCheckState("sortBySubjectMenuitem", (sortType == nsMsgViewSortType.bySubject));
     setSortByMenuItemCheckState("sortByUnreadMenuitem", (sortType == nsMsgViewSortType.byUnread));
-    setSortByMenuItemCheckState("sortByLabelMenuitem", (sortType == nsMsgViewSortType.byLabel));
+    setSortByMenuItemCheckState("sortByTagsMenuitem", (sortType == nsMsgViewSortType.byTags));
     setSortByMenuItemCheckState("sortByJunkStatusMenuitem", (sortType == nsMsgViewSortType.byJunkStatus));
     setSortByMenuItemCheckState("sortBySenderMenuitem", (sortType == nsMsgViewSortType.byAuthor));
     setSortByMenuItemCheckState("sortByRecipientMenuitem", (sortType == nsMsgViewSortType.byRecipient));
@@ -237,8 +238,9 @@ function InitViewSortByMenu()
     var sortOrder = gDBView.sortOrder;
     var sortTypeSupportsGrouping = (sortType == nsMsgViewSortType.byAuthor 
         || sortType == nsMsgViewSortType.byDate || sortType == nsMsgViewSortType.byPriority
-        || sortType == nsMsgViewSortType.bySubject || sortType == nsMsgViewSortType.byLabel
-        || sortType == nsMsgViewSortType.byRecipient);
+        || sortType == nsMsgViewSortType.bySubject || sortType == nsMsgViewSortType.byTags
+        || sortType == nsMsgViewSortType.byRecipient|| sortType == nsMsgViewSortType.byFlagged
+        || sortType == nsMsgViewSortType.byAttachments);
 
     setSortByMenuItemCheckState("sortAscending", (sortOrder == nsMsgViewSortOrder.ascending));
     setSortByMenuItemCheckState("sortDescending", (sortOrder == nsMsgViewSortOrder.descending));
@@ -286,76 +288,6 @@ function InitViewMessagesMenu()
     ignoredTheadsMenuItem.setAttribute("checked", (viewFlags & nsMsgViewFlagsType.kShowIgnored) != 0);
 }
 
-function InitViewMessageViewMenu()
-{
-  var currentViewValue = document.getElementById("viewPicker").value;
-
-  var allMenuItem = document.getElementById("viewAll");
-  if(allMenuItem)
-    allMenuItem.setAttribute("checked",  currentViewValue == 0);  // from msgViewPickerOveraly.xul <menuitem value="0" id="viewPickerAll" label="&viewPickerAll.label;"/>
-    
-  var unreadMenuItem = document.getElementById("viewUnread");
-  if(unreadMenuItem)
-    unreadMenuItem.setAttribute("checked", currentViewValue == 1); // from msgViewPickerOveraly.xul, <menuitem value="1" id="viewPickerUnread" label="&viewPickerUnread.label;"/>
-  
-  for (var i = 1; i <= 5; i++) {
-    var prefString = gPrefBranch.getComplexValue("mailnews.labels.description." + i, Components.interfaces.nsIPrefLocalizedString).data;
-    var viewLabelMenuItem = document.getElementById("viewLabelMenuItem" + i);
-    viewLabelMenuItem.setAttribute("label", prefString);
-    viewLabelMenuItem.setAttribute("checked", (i == (currentViewValue - 1)));  // 1=2-1, from msgViewPickerOveraly.xul, <menuitem value="2" id="labelMenuItem1"/>
-  }
-
-  viewRefreshCustomMailViews(currentViewValue);
-}
-
-function viewRefreshCustomMailViews(aCurrentViewValue)
-{
-  // for each mail view in the msg view list, add a menu item
-  var mailViewList = Components.classes["@mozilla.org/messenger/mailviewlist;1"].getService(Components.interfaces.nsIMsgMailViewList);
-
-  // XXX TODO, fix code in msgViewPickerOverlay.js, to be like this.
-  // remove any existing entries...
-  var menupopupNode = document.getElementById('viewMessageViewPopup');
-  var userDefinedItems = menupopupNode.getElementsByAttribute("userdefined","true");
-  for (var i=0; userDefinedItems.item(i); )
-  {
-    if (!menupopupNode.removeChild(userDefinedItems[i]))
-      ++i;
-  }
-  
-  // now rebuild the list
-  var numItems = mailViewList.mailViewCount; 
-  var viewCreateCustomViewSeparator = document.getElementById('viewCreateCustomViewSeparator');
-  
-  for (i = 0; i < numItems; i++)
-  {
-    var newMenuItem = document.createElement("menuitem");
-    newMenuItem.setAttribute("label", mailViewList.getMailViewAt(i).prettyName);
-    newMenuItem.setAttribute("userdefined", "true");
-    var oncommandStr = "ViewMessagesBy('userdefinedview" + (kLastDefaultViewIndex + i) + "');";  
-    newMenuItem.setAttribute("oncommand", oncommandStr);
-    var item = menupopupNode.insertBefore(newMenuItem, viewCreateCustomViewSeparator);
-    item.setAttribute("value",  kLastDefaultViewIndex + i); 
-    item.setAttribute("type",  "radio"); // for checked
-    item.setAttribute("name", "viewmessages");  // for checked
-    item.setAttribute("checked", (kLastDefaultViewIndex + i == aCurrentViewValue)); 
-  }
-
-  if (!numItems)
-    viewCreateCustomViewSeparator.setAttribute('collapsed', true);
-  else
-    viewCreateCustomViewSeparator.removeAttribute('collapsed');
-}
-
-// called by the View | Messages | Views ... menu items
-// see mailWindowOverlay.xul
-function ViewMessagesBy(id)
-{
-  var viewPicker = document.getElementById('viewPicker');
-  viewPicker.selectedItem = document.getElementById(id);
-  viewChange(viewPicker, viewPicker.value);
-}
-
 function InitMessageMenu()
 {
   var aMessage = GetFirstSelectedMessage();
@@ -412,14 +344,14 @@ function InitMessageMenu()
   if(copyMenu)
       copyMenu.setAttribute("disabled", !aMessage);
 
-  // Disable Forward as/Label menu items if no message is selected
+  // Disable Forward as/tag menu items if no message is selected
   var forwardAsMenu = document.getElementById("forwardAsMenu");
   if(forwardAsMenu)
       forwardAsMenu.setAttribute("disabled", !aMessage);
 
-  var labelMenu = document.getElementById("labelMenu");
-  if(labelMenu)
-      labelMenu.setAttribute("disabled", !aMessage);
+  var tagMenu = document.getElementById("tagMenu");
+  if(tagMenu)
+      tagMenu.setAttribute("disabled", !aMessage);
 
   // Disable mark menu when we're not in a folder
   var markMenu = document.getElementById("markMenu");
@@ -519,65 +451,155 @@ function IsImapMessage(messageUri)
 
 function SetMenuItemLabel(menuItemId, customLabel)
 {
-    var menuItem = document.getElementById(menuItemId);
-
-    if(menuItem)
-        menuItem.setAttribute('label', customLabel);
+  var menuItem = document.getElementById(menuItemId);
+  if (menuItem)
+    menuItem.setAttribute('label', customLabel);
 }
 
-function InitMessageLabel(menuType)
+function RemoveAllMessageTags()
 {
-    /* this code gets the label strings and changes the menu labels */
-    var color;
+  var selectedMsgUris = GetSelectedMessages();
+  if (!selectedMsgUris.length)
+    return;
+    
+  for (var i = 0; i < selectedMsgUris.length; ++i)
+  {
+    // remove all tags by removing all their tag keys at once
+    // (using a msgHdr's setStringProperty won't notify the threadPane!)
+    var msgHdr = messenger.msgHdrFromURI(selectedMsgUris[i]);
+    msgHdr.label = 0; // remove legacy label
+    msgHdr.folder.getMsgDatabase(msgWindow)
+          .setStringProperty(msgHdr.messageKey, "keywords", "");
+  }
+  OnTagsChange();
+}
 
-    try
+function ToggleMessageTagKey(index)
+{
+  if (GetNumSelectedMessages() < 1)
+    return;
+  // toggle the tag state based upon that of the first selected message
+  var msgHdr = gDBView.hdrForFirstSelectedMessage;
+  var tagService = Components.classes["@mozilla.org/messenger/tagservice;1"]
+                             .getService(Components.interfaces.nsIMsgTagService);
+  var tagArray = tagService.getAllTags({});
+  for (var i = 0; i < tagArray.length; ++i)
+  {
+    var key = tagArray[i].key;
+    if (!--index) 
     {
-        var isChecked = true;
-        var checkedLabel = gDBView.hdrForFirstSelectedMessage.label;
+      // found the key, now toggle its state
+      var curKeys = msgHdr.getStringProperty("keywords");
+      if (msgHdr.label)
+        curKeys += " $label" + msgHdr.label;
+      var addKey  = (" " + curKeys + " ").indexOf(" " + key + " ") < 0;
+      ToggleMessageTag(key, addKey);
+      return;
     }
-    catch(ex)
+  }
+}
+
+function ToggleMessageTagMenu(target)
+{
+  var key    = target.getAttribute("value");
+  var addKey = target.getAttribute("checked") == "true";
+  ToggleMessageTag(key, addKey);
+}
+
+function ToggleMessageTag(key, addKey)
+{
+  var messages = Components.classes["@mozilla.org/supports-array;1"]
+                           .createInstance(Components.interfaces.nsISupportsArray);
+  var msg = Components.classes["@mozilla.org/supports-array;1"]
+                          .createInstance(Components.interfaces.nsISupportsArray);
+  var selectedMsgUris = GetSelectedMessages();
+  var toggler = addKey ? "addKeywordToMessages" : "removeKeywordFromMessages";
+  var prevHdrFolder = null;
+  // this crudely handles cross-folder virtual folders with selected messages
+  // that spans folders, by coalescing consecutive msgs in the selection
+  // that happen to be in the same folder. nsMsgSearchDBView does this
+  // better, but nsIMsgDBView doesn't handle commands with arguments,
+  // and (un)tag takes a key argument.
+  for (var i = 0; i < selectedMsgUris.length; ++i)
+  {
+    var msgHdr = messenger.msgHdrFromURI(selectedMsgUris[i]);
+    if (msgHdr.label)
     {
-        isChecked = false;
+      // Since we touch all these messages anyway, migrate the label now.
+      // If we don't, the thread tree won't always show the correct tag state,
+      // because resetting a label doesn't update the tree anymore...
+      msg.Clear();
+      msg.AppendElement(msgHdr);
+      msgHdr.folder.addKeywordToMessages(msg, "$label" + msgHdr.label);
+      msgHdr.label = 0; // remove legacy label
     }
-
-    for (var label = 0; label <= 5; label++)
+    if (prevHdrFolder != msgHdr.folder)
     {
-        try
-        {
-            var prefString = gPrefBranch.getComplexValue("mailnews.labels.description." + label,
-                                                         Components.interfaces.nsIPrefLocalizedString);
-            var formattedPrefString = gMessengerBundle.getFormattedString("labelMenuItemFormat" + label,
-                                                                          [prefString], 1); 
-            var menuItemId = menuType + "-labelMenuItem" + label;
-            var menuItem = document.getElementById(menuItemId);
-
-            SetMenuItemLabel(menuItemId, formattedPrefString);
-            if (isChecked && label == checkedLabel)
-              menuItem.setAttribute("checked", "true");
-            else
-              menuItem.setAttribute("checked", "false");
-
-            // commented out for now until UE decides on how to show the Labels menu items.
-            // This code will color either the text or background for the Labels menu items.
-            /*****
-            if (label != 0)
-            {
-                color = gPrefBranch.getCharPref("mailnews.labels.color." + label);
-                // this colors the text of the menuitem only.
-                //menuItem.setAttribute("style", ("color: " + color));
-
-                // this colors the background of the menuitem and
-                // when selected, text becomes white.
-                //menuItem.setAttribute("style", ("color: #FFFFFF"));
-                //menuItem.setAttribute("style", ("background-color: " + color));
-            }
-            ****/
-        }
-        catch(ex)
-        {
-        }
+      if (prevHdrFolder)
+        prevHdrFolder[toggler](messages, key);
+      messages.Clear();
+      prevHdrFolder = msgHdr.folder;
     }
-    document.commandDispatcher.updateCommands('create-menu-label');
+    messages.AppendElement(msgHdr);
+  }
+  if (prevHdrFolder)
+    prevHdrFolder[toggler](messages, key);
+  OnTagsChange();
+}
+
+function SetMessageTagLabel(menuitem, index, name)
+{
+  // if a <key> is defined for this tag, use its key as the accesskey
+  // (the key for the tag at index n needs to have the id key_tag<n>)
+  var shortcutkey = document.getElementById("key_tag" + index);
+  var accesskey = shortcutkey ? shortcutkey.getAttribute("key") : "";
+  if (accesskey)
+    menuitem.setAttribute("accesskey", accesskey);
+  var label = gMessengerBundle.getFormattedString("mailnews.tags.format", 
+                                                  [accesskey, name]);
+  menuitem.setAttribute("label", label);
+}
+
+function InitMessageTags(menuPopup)
+{
+  var tagService = Components.classes["@mozilla.org/messenger/tagservice;1"]
+                             .getService(Components.interfaces.nsIMsgTagService);
+  var tagArray = tagService.getAllTags({});
+  var tagCount = tagArray.length;
+
+  // remove any existing non-static entries...
+  var menuseparator = menuPopup.lastChild.previousSibling;
+  for (var i = menuPopup.childNodes.length; i > 4; --i)
+    menuPopup.removeChild(menuseparator.previousSibling);
+
+  // hide double menuseparator
+  menuseparator.previousSibling.hidden = !tagCount;
+
+  // create label and accesskey for the static remove item
+  var tagRemoveLabel = gMessengerBundle.getString("mailnews.tags.remove");
+  SetMessageTagLabel(menuPopup.firstChild, 0, tagRemoveLabel);
+
+  // now rebuild the list
+  var msgHdr = gDBView.hdrForFirstSelectedMessage;
+  var curKeys = msgHdr.getStringProperty("keywords");
+  if (msgHdr.label)
+    curKeys += " $label" + msgHdr.label;
+  for (var i = 0; i < tagCount; ++i)
+  {
+    var taginfo = tagArray[i];
+    // TODO we want to either remove or "check" the tags that already exist
+    var newMenuItem = document.createElement("menuitem");
+    SetMessageTagLabel(newMenuItem, i + 1, taginfo.tag);
+    newMenuItem.setAttribute("value", taginfo.key);
+    newMenuItem.setAttribute("type", "checkbox");
+    var removeKey = (" " + curKeys + " ").indexOf(" " + taginfo.key + " ") > -1;
+    newMenuItem.setAttribute('checked', removeKey);
+    newMenuItem.setAttribute('oncommand', 'ToggleMessageTagMenu(event.target);');
+    var color = taginfo.color;
+    if (color)
+      newMenuItem.setAttribute("class", "lc-" + color.substr(1));
+    menuPopup.insertBefore(newMenuItem, menuseparator);
+  }
 }
 
 function InitMessageMark()
@@ -1163,7 +1185,7 @@ function MsgOpenFromFile()
   var uri = fp.fileURL;
   uri.query = "type=application/x-message-display";
 
-  MsgOpenNewWindowForMessage(uri, null);
+  window.openDialog( "chrome://messenger/content/messageWindow.xul", "_blank", "all,chrome,dialog=no,status,toolbar", uri, null, null );
 }
 
 function MsgOpenNewWindowForMsgHdr(hdr)
@@ -2005,7 +2027,7 @@ function GetMessagesForAccount(aEvent)
   var uri = aEvent.target.id;
   var server = GetServer(uri);
   GetMessagesForInboxOnServer(server);
-  aEvent.preventBubble();
+  aEvent.stopPropagation();
 }
 
 
@@ -2092,8 +2114,8 @@ var gMessageNotificationBar =
   // flag bit values for mBarStatus, indexed by kMsgNotificationXXX
   mBarFlagValues: [
                     0, // for no msgNotificationBar
-                    1, // 1 << (kMsgNotificationJunkBar - 1)
-                    2, // 1 << (kMsgNotificationPhishingBar - 1)
+                    1, // 1 << (kMsgNotificationPhishingBar - 1)
+                    2, // 1 << (kMsgNotificationJunkBar - 1)
                     4  // 1 << (kMsgNotificationRemoteImages - 1)
                   ],
 
@@ -2102,7 +2124,6 @@ var gMessageNotificationBar =
   setJunkMsg: function(aMsgHdr)
   {
     var isJunk = false;
-    var isCurrentlyNotJunk = this.mMsgNotificationBar.selectedIndex != kMsgNotificationJunkBar;
   
     if (aMsgHdr) 
     {
@@ -2117,9 +2138,13 @@ var gMessageNotificationBar =
 
   setRemoteContentMsg: function(aMsgHdr)
   {  
-    var blockRemote = aMsgHdr &&
-                      aMsgHdr.getUint32Property("remoteContentPolicy") == kBlockRemoteContent;
-    this.updateMsgNotificationBar(kMsgNotificationRemoteImages, blockRemote);
+    // update the allow remote content for sender string
+    var headerParser = Components.classes["@mozilla.org/messenger/headerparser;1"]
+                                 .getService(Components.interfaces.nsIMsgHeaderParser);
+    var emailAddress = headerParser.extractHeaderAddressMailboxes(null, aMsgHdr.author);
+    document.getElementById('allowRemoteContentForAuthorDesc').value =
+      gMessengerBundle.getFormattedString('alwaysLoadRemoteContentForSender', [emailAddress ? emailAddress : aMsgHdr.author]);
+    this.updateMsgNotificationBar(kMsgNotificationRemoteImages, true);
   },
 
   // aUrl is the nsIURI for the message currently loaded in the message pane
@@ -2147,7 +2172,7 @@ var gMessageNotificationBar =
     var status = aSet ? this.mBarStatus | chunk : this.mBarStatus & ~chunk;
     this.mBarStatus = status;
 
-    // the junk message takes precedence over the phishing message
+    // the phishing message takes precedence over the junk message
     // which takes precedence over the remote content message
     this.mMsgNotificationBar.selectedIndex = this.mBarFlagValues.indexOf(status & -status);
 
@@ -2155,7 +2180,11 @@ var gMessageNotificationBar =
   }
 };
 
-function LoadMsgWithRemoteContent()
+/**
+ * loadMsgWithRemoteContent
+ *   Reload the current message, allowing remote content
+ */
+function loadMsgWithRemoteContent()
 {
   // we want to get the msg hdr for the currently selected message
   // change the "remoteContentBar" property on it
@@ -2164,6 +2193,76 @@ function LoadMsgWithRemoteContent()
   setMsgHdrPropertyAndReload("remoteContentPolicy", kAllowRemoteContent);
 }
 
+/**
+ *  msgHdrForCurrentMessage
+ *   Returns the msg hdr associated with the current loaded message.
+ */
+function msgHdrForCurrentMessage()
+{
+  var msgURI = GetLoadedMessage();
+  return (msgURI && !(/type=application\/x-message-display/.test(msgURI))) ? messenger.msgHdrFromURI(msgURI) : null;
+}
+
+/**
+ *  Reloads the message after adjusting the remote content policy for the sender.
+ *  Iterate through the local address books looking for a card with the same e-mail address as the 
+ *  sender of the current loaded message. If we find a card, update the allow remote content field.
+ *  If we can't find a card, prompt the user with a new AB card dialog, pre-selecting the remote content field.
+ */
+function allowRemoteContentForSender()
+{
+  // get the sender of the msg hdr
+  var msgHdr = msgHdrForCurrentMessage();
+  if (!msgHdr)
+    return;
+
+  var headerParser = Components.classes["@mozilla.org/messenger/headerparser;1"]
+                               .getService(Components.interfaces.nsIMsgHeaderParser);
+  var names = {};
+  var addresses = {};
+  var fullNames = {};
+  var numAddresses;
+
+  numAddresses = headerParser.parseHeadersWithArray(msgHdr.author, addresses, names, fullNames);
+  var authorEmailAddress = addresses.value[0];
+  if (!authorEmailAddress)
+    return;
+
+  // search through all of our local address books looking for a match.
+  var parentDir = RDF.GetResource("moz-abdirectory://").QueryInterface(Components.interfaces.nsIAbDirectory);
+  var enumerator = parentDir.childNodes;
+  var cardForEmailAddress;
+  var addrbook;
+  while (!cardForEmailAddress && enumerator.hasMoreElements())
+  {
+    addrbook = enumerator.getNext();
+    if (addrbook instanceof Components.interfaces.nsIAbMDBDirectory)
+      cardForEmailAddress = addrbook.cardForEmailAddress(authorEmailAddress);
+  }
+
+  var allowRemoteContent = false;
+  if (cardForEmailAddress)
+  {
+    // set the property for remote content
+    cardForEmailAddress.allowRemoteContent = true;
+    cardForEmailAddress.editCardToDatabase("");
+    allowRemoteContent = true;
+  }
+  else
+  {
+    var args = {primaryEmail:authorEmailAddress, displayName:names.value[0],
+                allowRemoteContent:true};
+    // create a new card and set the property
+    window.openDialog("chrome://messenger/content/addressbook/abNewCardDialog.xul",
+                      "", "chrome,resizable=no,titlebar,modal,centerscreen", args);
+    allowRemoteContent = args.allowRemoteContent;
+  } 
+
+  // reload the message if we've updated the remote content policy for the sender  
+  if (allowRemoteContent)
+    MsgReload();
+}
+
 function MsgIsNotAScam()
 {
   // we want to get the msg hdr for the currently selected message
@@ -2178,16 +2277,11 @@ function setMsgHdrPropertyAndReload(aProperty, aValue)
   // we want to get the msg hdr for the currently selected message
   // change the appropiate property on it then reload the message
 
-  var msgURI = GetLoadedMessage();
-
-  if (msgURI && !(/type=application\/x-message-display/.test(msgURI)))
+  var msgHdr = msgHdrForCurrentMessage();
+  if (msgHdr)
   {
-    var msgHdr = messenger.msgHdrFromURI(msgURI);
-    if (msgHdr)
-    {
-      msgHdr.setUint32Property(aProperty, aValue);
-      MsgReload();
-    }
+    msgHdr.setUint32Property(aProperty, aValue);
+    MsgReload();
   }
 }
 
@@ -2196,14 +2290,8 @@ function checkMsgHdrPropertyIsNot(aProperty, aValue)
   // we want to get the msg hdr for the currently selected message,
   // get the appropiate property on it and then test against value.
 
-  var msgURI = GetLoadedMessage();
-    
-  if (msgURI && !(/type=application\/x-message-display/.test(msgURI)))
-  {
-    var msgHdr = messenger.msgHdrFromURI(msgURI);
-    return (msgHdr && msgHdr.getUint32Property(aProperty) != aValue);
-  }
-  return false;
+  var msgHdr = msgHdrForCurrentMessage();
+  return (msgHdr && msgHdr.getUint32Property(aProperty) != aValue);
 }
 
 function MarkCurrentMessageAsRead()
@@ -2230,9 +2318,13 @@ function OnMsgLoaded(aUrl)
     if (!aUrl)
       return;
 
-    var folder = aUrl.folder;
+    // nsIMsgMailNewsUrl.folder throws an error when opening .eml files.
+    var folder;
+    try {
+      folder = aUrl.folder;
+    } catch (ex) {}
+
     var msgURI = GetLoadedMessage();
-    var msgHdr = null;
 
     if (!folder || !msgURI)
       return;
@@ -2243,8 +2335,7 @@ function OnMsgLoaded(aUrl)
     // SetNextMessageAfterDelete() when the operation completes (bug 243532).
     gNextMessageViewIndexAfterDelete = -2;
 
-    if (!(/type=application\/x-message-display/.test(msgURI)))
-      msgHdr = messenger.msgHdrFromURI(msgURI);
+    var msgHdr = msgHdrForCurrentMessage();
 
     gMessageNotificationBar.setJunkMsg(msgHdr);
 
@@ -2253,7 +2344,7 @@ function OnMsgLoaded(aUrl)
     var markReadOnADelay = gPrefBranch.getBoolPref("mailnews.mark_message_read.delay");
     if (msgHdr && !msgHdr.isRead)
     {
-      var wintype = document.firstChild.getAttribute('windowtype');
+      var wintype = document.documentElement.getAttribute('windowtype');
       if (markReadOnADelay && wintype == "mail:3pane") // only use the timer if viewing using the 3-pane preview pane and the user has set the pref
         gMarkViewedMessageAsReadTimer = setTimeout(MarkCurrentMessageAsRead, gPrefBranch.getIntPref("mailnews.mark_message_read.delay.interval") * 1000);
       else
@@ -2316,12 +2407,9 @@ function HandleMDNResponse(aUrl)
 
   var msgFolder = aUrl.folder;
   var msgURI = GetLoadedMessage();
-  if (!msgFolder || !msgURI)
+  if (!msgFolder || !msgURI || IsNewsMessage(msgURI))
     return;
 
-	if (IsNewsMessage(msgURI))
-		return;
-
   // if the message is marked as junk, do NOT attempt to process a return receipt
   // in order to better protect the user
   if (SelectedMessagesAreJunk())
@@ -2346,7 +2434,7 @@ function HandleMDNResponse(aUrl)
     if (mimeMsgId)
       msgHdr.messageId = mimeMsgId;
   }
-  
+
   // After a msg is downloaded it's already marked READ at this point so we must check if
   // the msg has a "Disposition-Notification-To" header and no MDN report has been sent yet.
   var msgFlags = msgHdr.flags;
@@ -2357,7 +2445,7 @@ function HandleMDNResponse(aUrl)
   var oldDNTHeader = mimeHdr.extractHeader("Return-Receipt-To", false);
   if (!DNTHeader && !oldDNTHeader)
     return;
- 
+
   // Everything looks good so far, let's generate the MDN response.
   var mdnGenerator = Components.classes["@mozilla.org/messenger-mdn/generator;1"].
                                   createInstance(Components.interfaces.nsIMsgMdnGenerator);
@@ -2383,18 +2471,6 @@ function MsgSearchMessages()
   OpenOrFocusWindow(args, "mailnews:search", "chrome://messenger/content/SearchDialog.xul");
 }
 
-function MsgJunkMail()
-{
-  MsgJunkMailInfo(true);
-  var preselectedFolder = null;
-  if ("GetFirstSelectedMsgFolder" in window)
-    preselectedFolder = GetFirstSelectedMsgFolder();
-
-  var args = { folder: preselectedFolder };
-  window.openDialog("chrome://messenger/content/junkMail.xul", "",
-                    "chrome,resizable=no,modal,dialog", args);
-}
-
 function MsgJunkMailInfo(aCheckFirstUse)
 {
   if (aCheckFirstUse) {
