@@ -39,8 +39,13 @@
 
 #include "nsFormFillController.h"
 
+#ifdef MOZ_PLACES
+#include "nsStorageFormHistory.h"
+#include "nsIAutoCompleteSimpleResult.h"
+#else
 #include "nsFormHistory.h"
 #include "nsIAutoCompleteResultTypes.h"
+#endif
 #include "nsString.h"
 #include "nsReadableUtils.h"
 #include "nsIServiceManager.h"
@@ -75,6 +80,7 @@
 NS_INTERFACE_MAP_BEGIN(nsFormFillController)
   NS_INTERFACE_MAP_ENTRY(nsIFormFillController)
   NS_INTERFACE_MAP_ENTRY(nsIAutoCompleteInput)
+  NS_INTERFACE_MAP_ENTRY(nsIAutoCompleteInput_MOZILLA_1_8_BRANCH)
   NS_INTERFACE_MAP_ENTRY(nsIAutoCompleteSearch)
   NS_INTERFACE_MAP_ENTRY(nsIDOMFocusListener)
   NS_INTERFACE_MAP_ENTRY(nsIDOMKeyListener)
@@ -143,8 +149,8 @@ GetScreenOrigin(nsIDOMElement* aElement)
         scale = presContext->TwipsToPixels();
 
         nsIFrame* frame;
-        nsresult rv = presShell->GetPrimaryFrameFor(content, &frame);
-        if (NS_FAILED(rv))
+        presShell->GetPrimaryFrameFor(content, &frame);
+        if (!frame)
           return rect;
 
         nsIView* view;
@@ -240,6 +246,21 @@ nsFormFillController::SetPopupOpen(PRBool aPopupOpen)
 {
   if (mFocusedPopup) {
     if (aPopupOpen) {
+      // make sure input field is visible before showing popup (bug 320938)
+      nsCOMPtr<nsIContent> content = do_QueryInterface(mFocusedInput);
+      NS_ENSURE_STATE(content);
+      nsCOMPtr<nsIDocShell> docShell = GetDocShellForInput(mFocusedInput);
+      NS_ENSURE_STATE(docShell);
+      nsCOMPtr<nsIPresShell> presShell;
+      docShell->GetPresShell(getter_AddRefs(presShell));
+      NS_ENSURE_STATE(presShell);
+      nsIFrame *frame = nsnull;
+      presShell->GetPrimaryFrameFor(content.get(), &frame);
+      if (frame)
+        presShell->ScrollFrameIntoView(frame,
+                                       NS_PRESSHELL_SCROLL_IF_NOT_VISIBLE,
+                                       NS_PRESSHELL_SCROLL_IF_NOT_VISIBLE);
+
       nsRect popupRect = GetScreenOrigin(mFocusedInput);
       mFocusedPopup->OpenPopup(this, popupRect.x, popupRect.y+popupRect.height, popupRect.width);
     } else
@@ -394,16 +415,22 @@ nsFormFillController::GetSearchAt(PRUint32 index, nsACString & _retval)
 NS_IMETHODIMP
 nsFormFillController::GetTextValue(nsAString & aTextValue)
 {
-  mFocusedInput->GetValue(aTextValue);
+  if (mFocusedInput) {
+    mFocusedInput->GetValue(aTextValue);
+  } else {
+    aTextValue.Truncate();
+  }
   return NS_OK;
 }
 
 NS_IMETHODIMP
 nsFormFillController::SetTextValue(const nsAString & aTextValue)
 {
-  mSuppressOnInput = PR_TRUE;
-  mFocusedInput->SetValue(aTextValue);
-  mSuppressOnInput = PR_FALSE;
+  if (mFocusedInput) {
+    mSuppressOnInput = PR_TRUE;
+    mFocusedInput->SetValue(aTextValue);
+    mSuppressOnInput = PR_FALSE;
+  }
   return NS_OK;
 }
 
@@ -480,6 +507,16 @@ nsFormFillController::OnTextReverted(PRBool *_retval)
   return NS_OK;
 }
 
+////////////////////////////////////////////////////////////////////////
+//// nsIAutoCompleteInput_MOZILLA_1_8_BRANCH
+
+NS_IMETHODIMP
+nsFormFillController::GetConsumeRollupEvent(PRBool *aConsumeRollupEvent)
+{
+  *aConsumeRollupEvent = PR_FALSE;
+  return NS_OK;
+}
+
 ////////////////////////////////////////////////////////////////////////
 //// nsIAutoCompleteSearch
 
@@ -489,17 +526,25 @@ nsFormFillController::StartSearch(const nsAString &aSearchString, const nsAStrin
                                   nsIAutoCompleteResult *aPreviousResult, nsIAutoCompleteObserver *aListener)
 {
   nsCOMPtr<nsIAutoCompleteResult> result;
-  nsCOMPtr<nsIAutoCompleteMdbResult2> mdbResult = do_QueryInterface(aPreviousResult);
+
+#ifdef MOZ_PLACES
+  // This assumes that FormHistory uses nsIAutoCompleteSimpleResult,
+  // while PasswordManager does not.
+  nsCOMPtr<nsIAutoCompleteSimpleResult> historyResult;
+#else
+  nsCOMPtr<nsIAutoCompleteMdbResult2> historyResult;
+#endif
+  historyResult = do_QueryInterface(aPreviousResult);
 
   nsPasswordManager* passMgr = nsPasswordManager::GetInstance();
   if (!passMgr)
     return NS_ERROR_OUT_OF_MEMORY;
 
   // Only hand off a previous result to the password manager if it's
-  // a password manager result (i.e. not an nsIAutoCompleteMdbResult2).
+  // a password manager result (i.e. not an nsIAutoCompleteMdb/SimpleResult).
 
   if (!passMgr->AutoCompleteSearch(aSearchString,
-                                   mdbResult ? nsnull : aPreviousResult,
+                                   historyResult ? nsnull : aPreviousResult,
                                    mFocusedInput,
                                    getter_AddRefs(result)))
   {
@@ -507,9 +552,8 @@ nsFormFillController::StartSearch(const nsAString &aSearchString, const nsAStrin
     if (history) {
       history->AutoCompleteSearch(aSearchParam,
                                   aSearchString,
-                                  mdbResult,
+                                  historyResult,
                                   getter_AddRefs(result));
-      NS_RELEASE(history);
     }
   }
   NS_RELEASE(passMgr);
@@ -611,21 +655,26 @@ nsFormFillController::KeyPress(nsIDOMEvent* aEvent)
   keyEvent->GetKeyCode(&k);
   switch (k) {
   case nsIDOMKeyEvent::DOM_VK_DELETE:
+#ifndef XP_MACOSX
+    mController->HandleDelete(&cancel);
+    break;
+  case nsIDOMKeyEvent::DOM_VK_BACK_SPACE:
+    mController->HandleText(PR_FALSE);
+    break;
+#else
+  case nsIDOMKeyEvent::DOM_VK_BACK_SPACE:
     {
       PRBool isShift = PR_FALSE;
       keyEvent->GetShiftKey(&isShift);
 
-      if (isShift) {
+      if (isShift)
         mController->HandleDelete(&cancel);
+      else
+        mController->HandleText(PR_FALSE);
 
-        break;
-      }
-
-      // fall through
+      break;
     }
-  case nsIDOMKeyEvent::DOM_VK_BACK_SPACE:
-    mController->HandleText(PR_FALSE);
-    break;
+#endif
   case nsIDOMKeyEvent::DOM_VK_UP:
     mController->HandleKeyNavigation(nsIAutoCompleteController::KEY_UP, &cancel);
     break;
@@ -1073,7 +1122,7 @@ nsFormFillController::GetDocShellForInput(nsIDOMHTMLInputElement *aInput)
   nsCOMPtr<nsIDOMDocument> domDoc;
   aInput->GetOwnerDocument(getter_AddRefs(domDoc));
   nsCOMPtr<nsIDocument> doc = do_QueryInterface(domDoc);
-  
+  NS_ENSURE_TRUE(doc, nsnull);
   nsCOMPtr<nsIWebNavigation> webNav = do_GetInterface(doc->GetScriptGlobalObject());
   nsCOMPtr<nsIDocShell> docShell = do_QueryInterface(webNav);
   return docShell;
